/*
 * Jalview - A Sequence Alignment Editor and Viewer (Version 2.8.2)
 * Copyright (C) 2014 The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.api.SequenceStructureBinding;
import jalview.api.structures.JalviewStructureDisplayI;
import jalview.bin.Cache;
import jalview.datamodel.Alignment;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.ColumnSelection;
import jalview.datamodel.PDBEntry;
import jalview.datamodel.SequenceI;
import jalview.gui.ViewSelectionMenu.ViewSetProvider;
import jalview.io.AppletFormatAdapter;
import jalview.io.JalviewFileChooser;
import jalview.io.JalviewFileView;
import jalview.jbgui.GStructureViewer;
import jalview.schemes.BuriedColourScheme;
import jalview.schemes.ColourSchemeI;
import jalview.schemes.HelixColourScheme;
import jalview.schemes.HydrophobicColourScheme;
import jalview.schemes.PurinePyrimidineColourScheme;
import jalview.schemes.StrandColourScheme;
import jalview.schemes.TaylorColourScheme;
import jalview.schemes.TurnColourScheme;
import jalview.schemes.ZappoColourScheme;
import jalview.util.MessageManager;
import jalview.util.Platform;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JColorChooser;
import javax.swing.JInternalFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

/**
 * GUI elements for handlnig an external chimera display
 * 
 * @author jprocter
 *
 */
public class ChimeraViewFrame extends GStructureViewer implements Runnable,
        ViewSetProvider, JalviewStructureDisplayI

{
  JalviewChimeraBindingModel jmb;

  AlignmentPanel ap;

  Vector atomsPicked = new Vector();

  private boolean addingStructures = false;

  ViewSelectionMenu seqColourBy;

  /**
   * 
   * @param files
   * @param ids
   * @param seqs
   * @param ap
   * @param usetoColour
   *          - add the alignment panel to the list used for colouring these
   *          structures
   * @param useToAlign
   *          - add the alignment panel to the list used for aligning these
   *          structures
   * @param leaveColouringToJmol
   *          - do not update the colours from any other source. Jmol is
   *          handling them
   * @param loadStatus
   * @param bounds
   * @param viewid
   * 
   *          public ChimeraViewFrame(String[] files, String[] ids,
   *          SequenceI[][] seqs, AlignmentPanel ap, boolean usetoColour,
   *          boolean useToAlign, boolean leaveColouringToJmol, String
   *          loadStatus, Rectangle bounds, String viewid) { PDBEntry[]
   *          pdbentrys = new PDBEntry[files.length]; for (int i = 0; i <
   *          pdbentrys.length; i++) { PDBEntry pdbentry = new PDBEntry();
   *          pdbentry.setFile(files[i]); pdbentry.setId(ids[i]); pdbentrys[i] =
   *          pdbentry; } // / TODO: check if protocol is needed to be set, and
   *          if chains are // autodiscovered. jmb = new
   *          JalviewChimeraBindingModel(this,
   *          ap.getStructureSelectionManager(), pdbentrys, seqs, null, null);
   * 
   *          jmb.setLoadingFromArchive(true); addAlignmentPanel(ap); if
   *          (useToAlign) { useAlignmentPanelForSuperposition(ap); } if
   *          (leaveColouringToJmol || !usetoColour) {
   *          jmb.setColourBySequence(false); seqColour.setSelected(false);
   *          jmolColour.setSelected(true); } if (usetoColour) {
   *          useAlignmentPanelForColourbyseq(ap);
   *          jmb.setColourBySequence(true); seqColour.setSelected(true);
   *          jmolColour.setSelected(false); } this.setBounds(bounds);
   *          initMenus(); viewId = viewid; //
   *          jalview.gui.Desktop.addInternalFrame(this, "Loading File", //
   *          bounds.width,bounds.height);
   * 
   *          this.addInternalFrameListener(new InternalFrameAdapter() { public
   *          void internalFrameClosing(InternalFrameEvent internalFrameEvent) {
   *          closeViewer(); } }); initJmol(loadStatus); // pdbentry, seq,
   *          JBPCHECK!
   * 
   *          }
   */
  private void initMenus()
  {
    seqColour.setSelected(jmb.isColourBySequence());
    jmolColour.setSelected(!jmb.isColourBySequence());
    if (_colourwith == null)
    {
      _colourwith = new Vector<AlignmentPanel>();
    }
    if (_alignwith == null)
    {
      _alignwith = new Vector<AlignmentPanel>();
    }

    seqColourBy = new ViewSelectionMenu(MessageManager.getString("label.colour_by"), this, _colourwith,
            new ItemListener()
            {

              @Override
              public void itemStateChanged(ItemEvent e)
              {
                if (!seqColour.isSelected())
                {
                  seqColour.doClick();
                }
                else
                {
                  // update the jmol display now.
                  seqColour_actionPerformed(null);
                }
              }
            });
    viewMenu.add(seqColourBy);
    final ItemListener handler;
    JMenu alpanels = new ViewSelectionMenu(MessageManager.getString("label.superpose_with"), this,
            _alignwith, handler = new ItemListener()
            {

              @Override
              public void itemStateChanged(ItemEvent e)
              {
                alignStructs.setEnabled(_alignwith.size() > 0);
                alignStructs.setToolTipText(MessageManager
                        .formatMessage(
                                "label.align_structures_using_linked_alignment_views",
                                new String[]
                                { new Integer(_alignwith.size()).toString() }));
              }
            });
    handler.itemStateChanged(null);
    jmolActionMenu.add(alpanels);
    jmolActionMenu.addMenuListener(new MenuListener()
    {

      @Override
      public void menuSelected(MenuEvent e)
      {
        handler.itemStateChanged(null);
      }

      @Override
      public void menuDeselected(MenuEvent e)
      {
        // TODO Auto-generated method stub

      }

      @Override
      public void menuCanceled(MenuEvent e)
      {
        // TODO Auto-generated method stub

      }
    });
  }

  IProgressIndicator progressBar = null;

  /**
   * add a single PDB structure to a new or existing Jmol view
   * 
   * @param pdbentry
   * @param seq
   * @param chains
   * @param ap
   */
  public ChimeraViewFrame(PDBEntry pdbentry, SequenceI[] seq,
          String[] chains, final AlignmentPanel ap)
  {
    progressBar = ap.alignFrame;
    // ////////////////////////////////
    // Is the pdb file already loaded?
    String alreadyMapped = ap.getStructureSelectionManager()
            .alreadyMappedToFile(pdbentry.getId());

    if (alreadyMapped != null)
    {
      int option = JOptionPane.showInternalConfirmDialog(Desktop.desktop,
              MessageManager.formatMessage(
                      "label.pdb_entry_is_already_displayed", new String[]
                      { pdbentry.getId() }), MessageManager.formatMessage(
                      "label.map_sequences_to_visible_window", new String[]
                      { pdbentry.getId() }), JOptionPane.YES_NO_OPTION);

      if (option == JOptionPane.YES_OPTION)
      {
        // TODO : Fix multiple seq to one chain issue here.
        ap.getStructureSelectionManager().setMapping(seq, chains,
                alreadyMapped, AppletFormatAdapter.FILE);
        if (ap.seqPanel.seqCanvas.fr != null)
        {
          ap.seqPanel.seqCanvas.fr.featuresAdded();
          ap.paintAlignment(true);
        }

        // Now this AppJmol is mapped to new sequences. We must add them to
        // the exisiting array
        JInternalFrame[] frames = Desktop.instance.getAllFrames();

        for (int i = 0; i < frames.length; i++)
        {
          if (frames[i] instanceof ChimeraViewFrame)
          {
            final ChimeraViewFrame topJmol = ((ChimeraViewFrame) frames[i]);
            // JBPNOTE: this looks like a binding routine, rather than a gui
            // routine
            for (int pe = 0; pe < topJmol.jmb.pdbentry.length; pe++)
            {
              if (topJmol.jmb.pdbentry[pe].getFile().equals(alreadyMapped))
              {
                topJmol.jmb.addSequence(pe, seq);
                topJmol.addAlignmentPanel(ap);
                // add it to the set used for colouring
                topJmol.useAlignmentPanelForColourbyseq(ap);
                topJmol.buildChimeraActionMenu();
                ap.getStructureSelectionManager()
                        .sequenceColoursChanged(ap);
                break;
              }
            }
          }
        }

        return;
      }
    }
    // /////////////////////////////////
    // Check if there are other Jmol views involving this alignment
    // and prompt user about adding this molecule to one of them
    Vector existingViews = getJmolsFor(ap);
    if (existingViews.size() > 0)
    {
      Enumeration jm = existingViews.elements();
      while (jm.hasMoreElements())
      {
        ChimeraViewFrame topJmol = (ChimeraViewFrame) jm.nextElement();
        // TODO: highlight topJmol in view somehow
        int option = JOptionPane
                .showInternalConfirmDialog(
                        Desktop.desktop,
                        MessageManager.formatMessage(
                                "label.add_pdbentry_to_view", new String[]
                                { pdbentry.getId(), topJmol.getTitle() }),
                        MessageManager
                                .getString("label.align_to_existing_structure_view"),
                        JOptionPane.YES_NO_OPTION);
        if (option == JOptionPane.YES_OPTION)
        {
          topJmol.useAlignmentPanelForSuperposition(ap);
          topJmol.addStructure(pdbentry, seq, chains, true, ap.alignFrame);
          return;
        }
      }
    }
    // /////////////////////////////////
    openNewJmol(ap, new PDBEntry[]
    { pdbentry }, new SequenceI[][]
    { seq });
  }

  private void openNewJmol(AlignmentPanel ap, PDBEntry[] pdbentrys,
          SequenceI[][] seqs)
  {
    progressBar = ap.alignFrame;
    jmb = new JalviewChimeraBindingModel(this,
            ap.getStructureSelectionManager(), pdbentrys, seqs, null, null);
    addAlignmentPanel(ap);
    useAlignmentPanelForColourbyseq(ap);
    if (pdbentrys.length > 1)
    {
      alignAddedStructures = true;
      useAlignmentPanelForSuperposition(ap);
    }
    jmb.setColourBySequence(true);
    setSize(400, 400); // probably should be a configurable/dynamic default here
    initMenus();
    worker = null;
    {
      addingStructures = false;
      worker = new Thread(this);
      worker.start();
    }
    this.addInternalFrameListener(new InternalFrameAdapter()
    {
      public void internalFrameClosing(InternalFrameEvent internalFrameEvent)
      {
        closeViewer();
      }
    });

  }

  /**
   * create a new Jmol containing several structures superimposed using the
   * given alignPanel.
   * 
   * @param ap
   * @param pe
   * @param seqs
   */
  public ChimeraViewFrame(AlignmentPanel ap, PDBEntry[] pe,
          SequenceI[][] seqs)
  {
    openNewJmol(ap, pe, seqs);
  }

  /**
   * list of sequenceSet ids associated with the view
   */
  ArrayList<String> _aps = new ArrayList();

  public AlignmentPanel[] getAllAlignmentPanels()
  {
    AlignmentPanel[] t, list = new AlignmentPanel[0];
    for (String setid : _aps)
    {
      AlignmentPanel[] panels = PaintRefresher.getAssociatedPanels(setid);
      if (panels != null)
      {
        t = new AlignmentPanel[list.length + panels.length];
        System.arraycopy(list, 0, t, 0, list.length);
        System.arraycopy(panels, 0, t, list.length, panels.length);
        list = t;
      }
    }

    return list;
  }

  /**
   * list of alignment panels to use for superposition
   */
  Vector<AlignmentPanel> _alignwith = new Vector<AlignmentPanel>();

  /**
   * list of alignment panels that are used for colouring structures by aligned
   * sequences
   */
  Vector<AlignmentPanel> _colourwith = new Vector<AlignmentPanel>();

  /**
   * set the primary alignmentPanel reference and add another alignPanel to the
   * list of ones to use for colouring and aligning
   * 
   * @param nap
   */
  public void addAlignmentPanel(AlignmentPanel nap)
  {
    if (ap == null)
    {
      ap = nap;
    }
    if (!_aps.contains(nap.av.getSequenceSetId()))
    {
      _aps.add(nap.av.getSequenceSetId());
    }
  }

  /**
   * remove any references held to the given alignment panel
   * 
   * @param nap
   */
  public void removeAlignmentPanel(AlignmentPanel nap)
  {
    try
    {
      _alignwith.remove(nap);
      _colourwith.remove(nap);
      if (ap == nap)
      {
        ap = null;
        for (AlignmentPanel aps : getAllAlignmentPanels())
        {
          if (aps != nap)
          {
            ap = aps;
            break;
          }
        }
      }
    } catch (Exception ex)
    {
    }
    if (ap != null)
    {
      buildChimeraActionMenu();
    }
  }

  public void useAlignmentPanelForSuperposition(AlignmentPanel nap)
  {
    addAlignmentPanel(nap);
    if (!_alignwith.contains(nap))
    {
      _alignwith.add(nap);
    }
  }

  public void excludeAlignmentPanelForSuperposition(AlignmentPanel nap)
  {
    if (_alignwith.contains(nap))
    {
      _alignwith.remove(nap);
    }
  }

  public void useAlignmentPanelForColourbyseq(AlignmentPanel nap,
          boolean enableColourBySeq)
  {
    useAlignmentPanelForColourbyseq(nap);
    jmb.setColourBySequence(enableColourBySeq);
    seqColour.setSelected(enableColourBySeq);
    jmolColour.setSelected(!enableColourBySeq);
  }

  public void useAlignmentPanelForColourbyseq(AlignmentPanel nap)
  {
    addAlignmentPanel(nap);
    if (!_colourwith.contains(nap))
    {
      _colourwith.add(nap);
    }
  }

  public void excludeAlignmentPanelForColourbyseq(AlignmentPanel nap)
  {
    if (_colourwith.contains(nap))
    {
      _colourwith.remove(nap);
    }
  }

  /**
   * pdb retrieval thread.
   */
  private Thread worker = null;

  /**
   * add a new structure (with associated sequences and chains) to this viewer,
   * retrieving it if necessary first.
   * 
   * @param pdbentry
   * @param seq
   * @param chains
   * @param alignFrame
   * @param align
   *          if true, new structure(s) will be align using associated alignment
   */
  private void addStructure(final PDBEntry pdbentry, final SequenceI[] seq,
          final String[] chains, final boolean b,
          final IProgressIndicator alignFrame)
  {
    if (pdbentry.getFile() == null)
    {
      if (worker != null && worker.isAlive())
      {
        // a retrieval is in progress, wait around and add ourselves to the
        // queue.
        new Thread(new Runnable()
        {
          public void run()
          {
            while (worker != null && worker.isAlive() && _started)
            {
              try
              {
                Thread.sleep(100 + ((int) Math.random() * 100));

              } catch (Exception e)
              {
              }

            }
            // and call ourselves again.
            addStructure(pdbentry, seq, chains, b, alignFrame);
          }
        }).start();
        return;
      }
    }
    // otherwise, start adding the structure.
    jmb.addSequenceAndChain(new PDBEntry[]
    { pdbentry }, new SequenceI[][]
    { seq }, new String[][]
    { chains });
    addingStructures = true;
    _started = false;
    alignAddedStructures = b;
    progressBar = alignFrame; // visual indication happens on caller frame.
    (worker = new Thread(this)).start();
    return;
  }

  private Vector getJmolsFor(AlignmentPanel ap2)
  {
    Vector otherJmols = new Vector();
    // Now this AppJmol is mapped to new sequences. We must add them to
    // the exisiting array
    JInternalFrame[] frames = Desktop.instance.getAllFrames();

    for (int i = 0; i < frames.length; i++)
    {
      if (frames[i] instanceof ChimeraViewFrame)
      {
        ChimeraViewFrame topJmol = ((ChimeraViewFrame) frames[i]);
        if (topJmol.isLinkedWith(ap2))
        {
          otherJmols.addElement(topJmol);
        }
      }
    }
    return otherJmols;
  }

  void initChimera(String command)
  {
    jmb.setFinishedInit(false);
    // TODO: consider waiting until the structure/view is fully loaded before
    // displaying
    jalview.gui.Desktop.addInternalFrame(this, jmb.getViewerTitle(),
            getBounds().width, getBounds().height);
    if (command == null)
    {
      command = "";
    }
    jmb.evalStateCommand(command, false);
    jmb.setFinishedInit(true);
  }

  void setChainMenuItems(Vector chains)
  {
    chainMenu.removeAll();
    if (chains == null)
    {
      return;
    }
    JMenuItem menuItem = new JMenuItem(
            MessageManager.getString("label.all"));
    menuItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent evt)
      {
        allChainsSelected = true;
        for (int i = 0; i < chainMenu.getItemCount(); i++)
        {
          if (chainMenu.getItem(i) instanceof JCheckBoxMenuItem)
            ((JCheckBoxMenuItem) chainMenu.getItem(i)).setSelected(true);
        }
        centerViewer();
        allChainsSelected = false;
      }
    });

    chainMenu.add(menuItem);

    for (int c = 0; c < chains.size(); c++)
    {
      menuItem = new JCheckBoxMenuItem(chains.elementAt(c).toString(), true);
      menuItem.addItemListener(new ItemListener()
      {
        public void itemStateChanged(ItemEvent evt)
        {
          if (!allChainsSelected)
            centerViewer();
        }
      });

      chainMenu.add(menuItem);
    }
  }

  boolean allChainsSelected = false;

  private boolean alignAddedStructures = false;

  void centerViewer()
  {
    Vector toshow = new Vector();
    String lbl;
    int mlength, p, mnum;
    for (int i = 0; i < chainMenu.getItemCount(); i++)
    {
      if (chainMenu.getItem(i) instanceof JCheckBoxMenuItem)
      {
        JCheckBoxMenuItem item = (JCheckBoxMenuItem) chainMenu.getItem(i);
        if (item.isSelected())
        {
          toshow.addElement(item.getText());
        }
      }
    }
    jmb.centerViewer(toshow);
  }

  public void closeViewer()
  {
    jmb.closeViewer();
    ap = null;
    _aps.clear();
    _alignwith.clear();
    _colourwith.clear();
    // TODO: check for memory leaks where instance isn't finalised because jmb
    // holds a reference to the window
    jmb = null;
  }

  /**
   * state flag for PDB retrieval thread
   */
  private boolean _started = false;

  public void run()
  {
    _started = true;
    String pdbid = "";
    // todo - record which pdbids were successfuly imported.
    StringBuffer errormsgs = new StringBuffer(), files = new StringBuffer();
    List<String> fileToLoad=new ArrayList<String>();
    List<PDBEntry> filePDB = new ArrayList<PDBEntry>();
    List<Integer> filePDBpos =new ArrayList<Integer>();
    try
    {
      String[] curfiles = jmb.getPdbFile(); // files currently in viewer
      // TODO: replace with reference fetching/transfer code (validate PDBentry
      // as a DBRef?)
      jalview.ws.dbsources.Pdb pdbclient = new jalview.ws.dbsources.Pdb();
      for (int pi = 0; pi < jmb.pdbentry.length; pi++)
      {
        String file = null;
        if (jmb.pdbentry[pi].getFile()==null) 
        {
          // retrieve the pdb and store it locally
          AlignmentI pdbseq = null;
          pdbid = jmb.pdbentry[pi].getId();
          long hdl = pdbid.hashCode() - System.currentTimeMillis();
          if (progressBar != null)
          {
            progressBar.setProgressBar("Fetching PDB " + pdbid, hdl);
          }
          try
          {
            pdbseq = pdbclient.getSequenceRecords(pdbid = jmb.pdbentry[pi]
                    .getId());
          } catch (OutOfMemoryError oomerror)
          {
            new OOMWarning("Retrieving PDB id " + pdbid, oomerror);
          } catch (Exception ex)
          {
            ex.printStackTrace();
            errormsgs.append("'" + pdbid + "'");
          }
          if (progressBar != null)
          {
            progressBar.setProgressBar("Finished.", hdl);
          }
          if (pdbseq != null)
          {
            // just transfer the file name from the first sequence's first
            // PDBEntry
            file = new File(((PDBEntry) pdbseq.getSequenceAt(0).getPDBId()
                    .elementAt(0)).getFile()).getAbsolutePath();
            jmb.pdbentry[pi].setFile(file);

            files.append(" \"" + Platform.escapeString(file) + "\"");
          }
          else
          {
            errormsgs.append("'" + pdbid + "' ");
          }
        }
        else
        {
          file = new File(jmb.pdbentry[pi].getFile())
          .getAbsoluteFile().getPath();
          if (curfiles != null && curfiles.length > 0)
          {
            addingStructures = true; // already files loaded.
            for (int c = 0; c < curfiles.length; c++)
            {
              if (curfiles[c].equals(file))
              {
                file = null;
                break;
              }
            }
          }
          
          if (file != null)
          {
            fileToLoad.add(file);
            filePDB.add(jmb.pdbentry[pi]);
            filePDBpos.add(Integer.valueOf(pi));
            files.append(" \"" + Platform.escapeString(file) + "\"");
          }
        }
      }
    } catch (OutOfMemoryError oomerror)
    {
      new OOMWarning("Retrieving PDB files: " + pdbid, oomerror);
    } catch (Exception ex)
    {
      ex.printStackTrace();
      errormsgs.append("When retrieving pdbfiles : current was: '" + pdbid
              + "'");
    }
    if (errormsgs.length() > 0)
    {

      JOptionPane.showInternalMessageDialog(Desktop.desktop, MessageManager
              .formatMessage("label.pdb_entries_couldnt_be_retrieved",
                      new String[]
                      { errormsgs.toString() }), MessageManager
              .getString("label.couldnt_load_file"),
              JOptionPane.ERROR_MESSAGE);

    }
    long lastnotify = jmb.getLoadNotifiesHandled();
    if (files.length() > 0)
    {
      if (!addingStructures)
      {
        try
        {
          initChimera("");
        } catch (Exception ex)
        {
          Cache.log.error("Couldn't open Chimera viewer!", ex);
        }
      } 
      int num=-1;
      for (PDBEntry pe : filePDB)
      {
        num++;
        if (pe.getFile() != null)
        {
          try
          {
            int pos=filePDBpos.get(num).intValue();
            jmb.openFile(pe);
            jmb.addSequence(pos, jmb.sequence[pos]);
            File fl=new File(pe.getFile());
            String protocol = AppletFormatAdapter.URL;
            try
            {
              if (fl.exists())
                {
                  protocol = AppletFormatAdapter.FILE;
                }
              } catch (Exception e)
              {
              } catch (Error e)
              {
              }
              // Explicitly map to the filename used by Jmol ;
              jmb.ssm.setMapping(jmb.sequence[pos], null, pe.getFile(),
                      protocol);
              // pdbentry[pe].getFile(), protocol);
          } catch (OutOfMemoryError oomerror)
          {
            new OOMWarning(
                    "When trying to open and map structures from Chimera!",
                    oomerror);
          } catch (Exception ex)
          {
            Cache.log.error("Couldn't open " + pe.getFile()
                    + " in Chimera viewer!", ex);
          } finally
          {
            Cache.log.debug("File locations are " + files);
          }
        }
      }
      // jmb.getPdbFile();
      jmb.setFinishedInit(true);
      jmb.setLoadingFromArchive(false);
      
      // refresh the sequence colours for the new structure(s)
      for (AlignmentPanel ap : _colourwith)
      {
        jmb.updateColours(ap);
      }
      // do superposition if asked to
      if (alignAddedStructures)
      {
        new Thread(new Runnable()
        {
          public void run()
          {
            alignStructs_withAllAlignPanels();
          }
        }).start();
        alignAddedStructures = false;
      }
      addingStructures = false;
    }
    _started = false;
    worker = null;
  }

  public void pdbFile_actionPerformed(ActionEvent actionEvent)
  {
    JalviewFileChooser chooser = new JalviewFileChooser(
            jalview.bin.Cache.getProperty("LAST_DIRECTORY"));

    chooser.setFileView(new JalviewFileView());
    chooser.setDialogTitle("Save PDB File");
    chooser.setToolTipText(MessageManager.getString("action.save"));

    int value = chooser.showSaveDialog(this);

    if (value == JalviewFileChooser.APPROVE_OPTION)
    {
      try
      {
        // TODO: cope with multiple PDB files in view
        BufferedReader in = new BufferedReader(new FileReader(
                jmb.getPdbFile()[0]));
        File outFile = chooser.getSelectedFile();

        PrintWriter out = new PrintWriter(new FileOutputStream(outFile));
        String data;
        while ((data = in.readLine()) != null)
        {
          if (!(data.indexOf("<PRE>") > -1 || data.indexOf("</PRE>") > -1))
          {
            out.println(data);
          }
        }
        out.close();
      } catch (Exception ex)
      {
        ex.printStackTrace();
      }
    }
  }

  public void viewMapping_actionPerformed(ActionEvent actionEvent)
  {
    jalview.gui.CutAndPasteTransfer cap = new jalview.gui.CutAndPasteTransfer();
    try
    {
      for (int pdbe = 0; pdbe < jmb.pdbentry.length; pdbe++)
      {
        cap.appendText(jmb.printMapping(jmb.pdbentry[pdbe].getFile()));
        cap.appendText("\n");
      }
    } catch (OutOfMemoryError e)
    {
      new OOMWarning(
              "composing sequence-structure alignments for display in text box.",
              e);
      cap.dispose();
      return;
    }
    jalview.gui.Desktop.addInternalFrame(cap,
            MessageManager.getString("label.pdb_sequence_mapping"), 550,
            600);
  }

  /**
   * DOCUMENT ME!
   * 
   * @param e
   *          DOCUMENT ME!
   */
  public void eps_actionPerformed(ActionEvent e)
  {
    throw new Error("EPS Generation not yet implemented.");
  }

  /**
   * DOCUMENT ME!
   * 
   * @param e
   *          DOCUMENT ME!
   */
  public void png_actionPerformed(ActionEvent e)
  {
    throw new Error("PNG Generation not yet implemented.");
  }

  public void jmolColour_actionPerformed(ActionEvent actionEvent)
  {
    if (jmolColour.isSelected())
    {
      // disable automatic sequence colouring.
      jmb.setColourBySequence(false);
    }
  }

  public void seqColour_actionPerformed(ActionEvent actionEvent)
  {
    jmb.setColourBySequence(seqColour.isSelected());
    if (_colourwith == null)
    {
      _colourwith = new Vector<AlignmentPanel>();
    }
    if (jmb.isColourBySequence())
    {
      if (!jmb.isLoadingFromArchive())
      {
        if (_colourwith.size() == 0 && ap != null)
        {
          // Make the currently displayed alignment panel the associated view
          _colourwith.add(ap.alignFrame.alignPanel);
        }
      }
      // Set the colour using the current view for the associated alignframe
      for (AlignmentPanel ap : _colourwith)
      {
        jmb.colourBySequence(ap.av.showSequenceFeatures, ap);
      }
    }
  }

  public void chainColour_actionPerformed(ActionEvent actionEvent)
  {
    chainColour.setSelected(true);
    jmb.colourByChain();
  }

  public void chargeColour_actionPerformed(ActionEvent actionEvent)
  {
    chargeColour.setSelected(true);
    jmb.colourByCharge();
  }

  public void zappoColour_actionPerformed(ActionEvent actionEvent)
  {
    zappoColour.setSelected(true);
    jmb.setJalviewColourScheme(new ZappoColourScheme());
  }

  public void taylorColour_actionPerformed(ActionEvent actionEvent)
  {
    taylorColour.setSelected(true);
    jmb.setJalviewColourScheme(new TaylorColourScheme());
  }

  public void hydroColour_actionPerformed(ActionEvent actionEvent)
  {
    hydroColour.setSelected(true);
    jmb.setJalviewColourScheme(new HydrophobicColourScheme());
  }

  public void helixColour_actionPerformed(ActionEvent actionEvent)
  {
    helixColour.setSelected(true);
    jmb.setJalviewColourScheme(new HelixColourScheme());
  }

  public void strandColour_actionPerformed(ActionEvent actionEvent)
  {
    strandColour.setSelected(true);
    jmb.setJalviewColourScheme(new StrandColourScheme());
  }

  public void turnColour_actionPerformed(ActionEvent actionEvent)
  {
    turnColour.setSelected(true);
    jmb.setJalviewColourScheme(new TurnColourScheme());
  }

  public void buriedColour_actionPerformed(ActionEvent actionEvent)
  {
    buriedColour.setSelected(true);
    jmb.setJalviewColourScheme(new BuriedColourScheme());
  }

  public void purinePyrimidineColour_actionPerformed(ActionEvent actionEvent)
  {
    setJalviewColourScheme(new PurinePyrimidineColourScheme());
  }

  public void userColour_actionPerformed(ActionEvent actionEvent)
  {
    userColour.setSelected(true);
    new UserDefinedColours(this, null);
  }

  public void backGround_actionPerformed(ActionEvent actionEvent)
  {
    java.awt.Color col = JColorChooser.showDialog(this,
            "Select Background Colour", null);
    if (col != null)
    {
      jmb.setBackgroundColour(col);
    }
  }

  public void jmolHelp_actionPerformed(ActionEvent actionEvent)
  {
    try
    {
      jalview.util.BrowserLauncher
              .openURL("https://www.cgl.ucsf.edu/chimera/docs/UsersGuide");
    } catch (Exception ex)
    {
    }
  }

  String viewId = null;

  public String getViewId()
  {
    if (viewId == null)
    {
      viewId = System.currentTimeMillis() + "." + this.hashCode();
    }
    return viewId;
  }

  public void updateTitleAndMenus()
  {
    if (jmb.fileLoadingError != null && jmb.fileLoadingError.length() > 0)
    {
      repaint();
      return;
    }
    setChainMenuItems(jmb.chainNames);

    this.setTitle(jmb.getViewerTitle());
    if (jmb.getPdbFile().length > 1 && jmb.sequence.length > 1)
    {
      jmolActionMenu.setVisible(true);
    }
    if (!jmb.isLoadingFromArchive())
    {
      seqColour_actionPerformed(null);
    }
  }

  protected void buildChimeraActionMenu()
  {
    if (_alignwith == null)
    {
      _alignwith = new Vector<AlignmentPanel>();
    }
    if (_alignwith.size() == 0 && ap != null)
    {
      _alignwith.add(ap);
    }
    ;
    for (Component c : jmolActionMenu.getMenuComponents())
    {
      if (c != alignStructs)
      {
        jmolActionMenu.remove((JMenuItem) c);
      }
    }
    final ItemListener handler;
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * jalview.jbgui.GStructureViewer#alignStructs_actionPerformed(java.awt.event
   * .ActionEvent)
   */
  @Override
  protected void alignStructs_actionPerformed(ActionEvent actionEvent)
  {
    alignStructs_withAllAlignPanels();
  }

  private void alignStructs_withAllAlignPanels()
  {
    if (ap == null)
    {
      return;
    }
    ;
    if (_alignwith.size() == 0)
    {
      _alignwith.add(ap);
    }
    ;
    try
    {
      AlignmentI[] als = new Alignment[_alignwith.size()];
      ColumnSelection[] alc = new ColumnSelection[_alignwith.size()];
      int[] alm = new int[_alignwith.size()];
      int a = 0;

      for (AlignmentPanel ap : _alignwith)
      {
        als[a] = ap.av.getAlignment();
        alm[a] = -1;
        alc[a++] = ap.av.getColumnSelection();
      }
      jmb.superposeStructures(als, alm, alc);
    } catch (Exception e)
    {
      StringBuffer sp = new StringBuffer();
      for (AlignmentPanel ap : _alignwith)
      {
        sp.append("'" + ap.alignFrame.getTitle() + "' ");
      }
      Cache.log.info("Couldn't align structures with the " + sp.toString()
              + "associated alignment panels.", e);

    }

  }

  public void setJalviewColourScheme(ColourSchemeI ucs)
  {
    jmb.setJalviewColourScheme(ucs);

  }

  /**
   * 
   * @param alignment
   * @return first alignment panel displaying given alignment, or the default
   *         alignment panel
   */
  public AlignmentPanel getAlignmentPanelFor(AlignmentI alignment)
  {
    for (AlignmentPanel ap : getAllAlignmentPanels())
    {
      if (ap.av.getAlignment() == alignment)
      {
        return ap;
      }
    }
    return ap;
  }

  /**
   * 
   * @param ap2
   * @return true if this Jmol instance is linked with the given alignPanel
   */
  public boolean isLinkedWith(AlignmentPanel ap2)
  {
    return _aps.contains(ap2.av.getSequenceSetId());
  }

  public boolean isUsedforaligment(AlignmentPanel ap2)
  {

    return (_alignwith != null) && _alignwith.contains(ap2);
  }

  public boolean isUsedforcolourby(AlignmentPanel ap2)
  {
    return (_colourwith != null) && _colourwith.contains(ap2);
  }

  /**
   * 
   * @return TRUE if the view is NOT being coloured by sequence associations.
   */
  public boolean isColouredByJmol()
  {
    return !jmb.isColourBySequence();
  }

  public SequenceStructureBinding getBinding()
  {
    return jmb;
  }

}
